###########################################################################################
# STLFilt.pl: STL error message decryption filter (A Perl script) supporting
#			  Microsoft VC++6 when working with any of the following libraries:
#					Native MSVC 6 Library
#					Dinkumware 3.08 Library
#					STLPort 4.x Library
#
# Copyright (C) 2001 by Leor Zolman (leor@bdsoft.com, www.bdsoft.com)
# Permission is granted to use this code without restriction
# provided this copyright notice is included in the code.
#
# Release 1.1: 9/02/2001
#
# For quick installation instructions for the STL Error Decryptor, see QUICKSTART.txt.
# For general info, see README.txt
#
# (Note: hard tab setting for this source file is: 4)
#
# Purpose:	Transform STL template-based typenames from MSVC error messages
#			into a minimal, *readable* format. We might lose some details...
#			but retain our sanity! For use with the "Proxy CL" program, CL.cpp.
#
# Command line arguments:
#			/iter:x (or) -iter:x         Set iterator policy to x,
#										 where x is s[mall], m[edium] or l[ong]
#										 (typically conveyed by the proxy CL.EXE)
#
# This file is part of a package including:
#	CL.cpp			The "Proxy" CL.EXE replacement,
#	STLTask			A taskbar icon-based error decryption control utility,
#	STLFilt.BAT		A "low-bandwidth" batch file for error decryption control
#
# Acknowledgments:
# -----------------
# Thanks to Dave Smallberg for the idea (wish I'd have thought of it myself)
# and to Scott Meyers for putting on his ESTL seminar -- the event that
# inspired this project. 
###########################################################################################

###########################################################################################
# User-configurable settings:
#

$newiter = 'iter';			# ('iter' or 'IT' or...) shorten the word "iterator" to this

							# default iterator policy:
$def_iter_policy = 'M';		# 'L' (Long): NEVER remove iterator type name qualification
							# 'M' (Medium): USUALLY remove iterator type name qualification
							#               leave intact when iter type may be significant
							#				to the diagnostic
							# 'S' (Short): ALWAYS remove iterator type name qualification

$alloc = 'alloc';			# shorten the word "allocator" to this
#
# END of user-configurable options
###########################################################################################

$| = 1;											# force output buffer flush after every line

$iter_policy = $def_iter_policy;				# default iterator policy

while ("$ARGV[0]")
{
	if ($ARGV[0] =~ /^[\/-]iter:([SML])[a-zA-Z]*$/i) # allow command-line iter_policy
	{												 # specification of form: /iter:x
		$iter_policy = "\u$1";
		shift;
		next;
	}

	last;
}

$id = '\b[a-zA-Z_]\w*';							# pattern for a simple identifier or keyword
$p = '(?: \*){0,5}';							# suffix for "ptr", "ptr to ptr", "ptr to ptr to ptr", ad nauseum.
$idp = "(?:$id )?(?:$id )?(?:$id )?$id ?$p ?";	# one or more identifiers/keywords with perhaps some *'s after

												# simple id or basic template spec
$cid = "(?:$idp(?: const)?|$id<$idp(?: const)?(?:,$idp(?: const)?){0,8}>$p) ?";

												# a cid or template type with 1+ cid's as parameters
$t = "(?:$cid|$id<$cid(?:,$cid){0,8}>$p|$id<$id<$cid>$p(?:,$id<$cid>$p){0,8} ?>$p)";

$long_id = 0;									# flag that remembers if last message processed 
												# was a long identifier warning
while (<>)
{
# throw away those obnoxious "long identifier" warnings, and subsequent references
# to those warnings (eliminates the need for the #pragma hack):
  if (/truncated to '255'/)
  {
	$long_id = 1;								# on long identifier warnings, set flag and cycle
	next;
  }
  if ($long_id == 1)
  {
	/see reference to class template/ && next;	# one kind of message referring to the warnings
	/\(\d+\) \: while/ && next;					# VC bug spews sentence fragments relating to those warnings
  }
  $long_id = 0;									# hopefully we've now caught all the crud that followed the long identifier warning... 

# eliminate some noise:
  s/\bstd\:\://g;
  s/\bclass //g;
  s/\bstruct ([^'])/\1/g;		# don't strip "struct" for *anonymous* structs
  s/\b__thiscall //g;
  s/\(__cdecl \*\)//g;
  s/\b__cdecl\W//g;
  s/\b_STLD?\:\://g;			# for STLPort

# simplify the ubiquitous "vanilla" string, cin, cout (w/optional default allocator):
  s/\b_?basic_(string|if?stream|of?stream|([io]?stringstream))<char,char_traits<char>(,_?allocator<char>)? ?>/\1/g;
  s/\b_?basic_(string|istream|ostream)<unsigned short,char_traits<unsigned short>(,_?allocator<unsigned short>)? ?>/w\1/g;
  s/\b_?basic_(string|istream|ostream)<_(E(lem)?|CharT),_Tr(aits)?,_A(lloc)?>/\1/g;


# The following loop repeats until no transformations occur in the last complete iteration:

  for ($pass = 1; ;$pass++)			# pass count (for debugging purposes only)
  {
	my $before = $_;				# save the current line; keep looping while changes happen

	s/\b([io])stream_iterator<($t),char,char_traits<char> *(,$t)?>/\1stream_iterator<\2>/g;

	# delete allocators from template typenames completely:
	s/,? ?allocator<$t ?> ?>/>/g;
	s/,? ?allocator<$t ?>(,[01]) ?>/\1>/g;

	# delete the allocator specification from parameter lists, *unless* the message
	# mentions an error converting *to* the allocator type (OK, OK, this is ugly... but the
	# only time we really want to think about the allocator is when an error message
	# refers to it...)

	unless (/' to '.*allocator/)
	{
	  s/,allocator<$t ?> ?//g;				# the leading comma allows the full spec.
	  s/,const allocator<$t ?> ?&//g;		# to appear in the error message details
	}

	# reduce any remaining allocator parameters to just "$alloc":
	s/\ballocator<$t ?> ?([\)'])/$alloc\1/g;
	s/\bconst allocator<$t ?> &([\)'])/const $alloc\1/g;


# This, BTW, is how I debug this thing...stick one of these before suspect code to see progressive x'formations
# (The "list" below would be replaced by a known pattern in the construct i'm debugging)
#/list/ && print "-----(Pass $pass)------------------------DEBUG:\n$_\n";

	# STLPort list and deque iterators:
	s/\b_List_iterator<($t),_Nonconst_traits<\1 ?> ?>/list<\1>::iterator/g;
	s/\b_List_iterator<($t),_Const_traits<\1 ?> ?>/list<\1>::const_iterator/g;
	s/\b_Deque_iterator<($t),_Nonconst_traits<\1 ?> ?>/deque<\1>::iterator/g;
	s/\b_Deque_iterator<($t),_Const_traits<\1 ?> ?>/deque<\1>::const_iterator/g;

	# STLPort set/multiset iterators:
	# Since STLPort uses the same iterator type for both set and multiset, we just
	# say "gen_set<T>::iterator" to mean the "GENERIC" set/multiset iterator type:
	s/\b_Rb_tree_iterator<($t),_Nonconst_traits<\1 ?> ?>/gen_set<\1>::iterator/g;
	s/\b_Rb_tree_iterator<($t),_Const_traits<\1 ?> ?>/gen_set<\1>::const_iterator/g;

	# STLPort map/multimap iterators (same "GENERIC" iterator approach as above):
	s/\b_Rb_tree_iterator<pair<($t) const ?,($t)>,_Nonconst_traits<pair<\1 const ?,\2> > >/gen_map<\1,\2>::iterator/g;
	s/\b_Rb_tree_iterator<pair<($t) const ?,($t)>,_Const_traits<pair<\1 const ?,\2> > >/gen_map<\1,\2>::const_iterator/g;

	# STLPort debug iterator: reduce the entire thing to just "dbg_iter"
	while (/\b_DBG_iter</)
	{
		$start = $pos = index ($_, "_DBG_iter<") + 10;
		$depth = 1;
		while ($depth)
		{
			$next = substr ($_, $pos++, 1);
			$depth++ if $next eq "<";
			$depth-- if $next eq ">";
		}
		$_ = substr($_, 0, $start - 1) . " " . substr($_, $pos);
	}
	s/\b_DBG_iter\b/dbg_iter/g;

	# Dinkumware map/multimap:
	s/\b(multi)?map<($t),($t)(,$t(<\2 ?>)?)? ?>/\1map<\2,\3\4>/g;
	s/\b_Tree<_Tmap_traits<($t),($t)(,$t(<\1 ?>)?)?,0 ?> ?>/map<\1,\2\3>/g;
	s/\b_Tree<_Tmap_traits<($t),($t)(,$t(<\1 ?>)?)?,1 ?> ?>/multimap<\1,\2\3>/g;

	# native VC++ map/multimap:
	s/\b_Tree<($t),pair<\1 const ?,($t) ?>,(multi)?map<\1,\2(,$t(<\1 ?>)?)? ?>\:\:_Kfn\4? ?>/\3map<\1,\2>/g;
	s/\(const pair<($t) const ?,($t) ?> \*,const pair<\1 const ?,\2 ?> \*,const $t<\1 ?> & ?&\)/($newiter, $newiter, \3<\1>)/g;

	# map/multimap members common to native/Dinkum:
	s/\bconst pair<($t) const ?,($t) ?> ([\*&])/pair<\1,\2> \3/g;

	# native VC++ set/multiset
 	s/\b_Tree<($t),\1,(multi)?set<\1(,$t(?:<\1 ?>)?)? ?(,$t)?>\:\:_Kfn(?:\3(?:<\1 ?>)?)? ?(,\4)?>/\2set<\1\3>/g;

	# Dinkumware set/multiset:
	s/\b_Tree<_Tset_traits<($t)(,$t)?,0 ?> ?>/set<\1\2>/g;
	s/\b_Tree<_Tset_traits<($t)(,$t)?,1 ?> ?>/multiset<\1\2>/g;

	# simplify default, ubiquitous "less" function objects (not "functors", right Chuck Allison?), but leave others intact:
	s/,_?less<$t ?>//g;
	s/,Comp<$t ?>//g;		# STLPort's default comparison function
	s/\b(,)?const less<$t ?> &/\1less &/g;

	last if $before eq $_;					# stop looping only if there were no changes this pass
  }

# reduce iterators according to $iter_policy:

  if ($iter_policy eq 'M')					# policy 'M': USUALLY remove:
  {
	unless (/( of type|' to '|from ')$t\:\:(const_)?iterator/  # Shorten to $newiter and
			|| /iterator' does/)							   # *remove* the base type completely...
	{														   # as long as the error message doesn't	
		s/$t\:\:(const_)?iterator\b/\1$newiter/g;			   # mention iterators!				
	}			
  }
  elsif ($iter_policy eq 'S')				# policy 'S': ALWAYS remove:
  {
	s/$t\:\:(const_)?iterator\b/\1$newiter/g;			#	remove the base type completely
  }
											# All policies (including 'L'):
  s/[^a-zA-Z0-9_']iterator\b/$newiter/g;	#	shorten "iterator" to "$newiter", 


# reduce "double" constructor names 'T::T' to just 'T':

  s/'string\:\:string([\('])/'string\1/g;
  s/'(.*)\:\:\1([\('])/'\1\2/g;

# I'm sorry, Microsoft needs grammar lessons:
  s/take 1 parameters/take 1 parameter/g;


# get rid of that useless space between stars in ptrs-to-ptrs:
  s/ \* \* \* \*/ ****/g;
  s/ \* \* \*/ ***/g;
  s/ \* \*/ **/g;

# deal with some other non-critical (and often not even very aesthetic) spaces:
  s/ >/>/g;
  s/ ,/,/g;

# and FINALLY, print out the result of all transformations:

  print;

}

exit 0;			# The proxy CL.EXE will take care of returning the appropriate status to caller

